"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse
import logging

import torch
import yaml


def convert_fine_tune_checkpoint(
    fine_tune_checkpoint_fn,
    output_checkpoint_fn,
    fine_tune_yaml_fn=None,
    output_yaml_fn=None,
):
    fine_tune_checkpoint = torch.load(
        fine_tune_checkpoint_fn, map_location="cpu", weights_only=False
    )

    if "config" not in fine_tune_checkpoint:
        raise KeyError("Finetune checkpoint does not have a valid 'config' field")

    try:
        starting_checkpoint_fn = fine_tune_checkpoint["config"]["model"][
            "finetune_config"
        ]["starting_checkpoint"]
    except KeyError as e:
        logging.error(
            f"Finetune config missing entry config/model/finetune_config/starting_checkpoint {fine_tune_checkpoint['config']}"
        )
        raise e

    starting_checkpoint = torch.load(
        starting_checkpoint_fn, map_location="cpu", weights_only=False
    )
    start_checkpoint_model_config = starting_checkpoint["config"]["model"]

    fine_tune_checkpoint["config"]["model"]["backbone"] = start_checkpoint_model_config[
        "backbone"
    ]
    # if we are data only, then copy over the heads config too
    ft_data_only = "heads" not in fine_tune_checkpoint["config"]["model"]
    if ft_data_only:
        fine_tune_checkpoint["config"]["model"]["heads"] = (
            start_checkpoint_model_config["heads"]
        )

    fine_tune_checkpoint["config"]["model"].pop("finetune_config")

    torch.save(fine_tune_checkpoint, output_checkpoint_fn)

    if fine_tune_yaml_fn is not None:
        with open(fine_tune_yaml_fn) as yaml_f:
            fine_tune_yaml = yaml.safe_load(yaml_f)
        fine_tune_yaml["model"].pop("finetune_config")
        fine_tune_yaml["model"]["backbone"] = start_checkpoint_model_config["backbone"]
        if ft_data_only:
            fine_tune_yaml["model"]["heads"] = start_checkpoint_model_config["heads"]
        with open(output_yaml_fn, "w") as yaml_file:
            yaml.dump(fine_tune_yaml, yaml_file)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--fine-tune-checkpoint",
        help="path to fine tuned checkpoint",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--output-release-checkpoint",
        help="path to output checkpoint",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--fine-tune-yaml",
        help="path to fine tune yaml config",
        type=str,
        required=False,
        default=None,
    )
    parser.add_argument(
        "--output-release-yaml",
        help="path to output yaml config",
        type=str,
        required=False,
        default=None,
    )
    args = parser.parse_args()

    convert_fine_tune_checkpoint(
        fine_tune_yaml_fn=args.fine_tune_yaml,
        fine_tune_checkpoint_fn=args.fine_tune_checkpoint,
        output_checkpoint_fn=args.output_release_checkpoint,
        output_yaml_fn=args.output_release_yaml,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import torch
import torch.nn as nn

if TYPE_CHECKING:
    from fairchem.core.models.escaip.configs import (
        GlobalConfigs,
        GraphNeuralNetworksConfigs,
        RegularizationConfigs,
    )
from fairchem.core.models.escaip.utils.nn_utils import (
    Activation,
    NormalizationType,
    get_feedforward,
    get_linear,
    get_normalization_layer,
)


class OutputProjection(nn.Module):
    def __init__(
        self,
        global_cfg: GlobalConfigs,
        gnn_cfg: GraphNeuralNetworksConfigs,
        reg_cfg: RegularizationConfigs,
    ):
        super().__init__()
        self.use_edge_readout = global_cfg.regress_forces and global_cfg.direct_forces
        self.use_global_readout = gnn_cfg.use_global_readout

        # map concatenated readout features to hidden size
        self.node_projection = get_linear(
            in_features=global_cfg.hidden_size * (global_cfg.num_layers + 1),
            out_features=global_cfg.hidden_size,
            activation=Activation(global_cfg.activation),
            bias=True,
        )
        self.output_norm_node = get_normalization_layer(
            NormalizationType(reg_cfg.normalization)
        )(global_cfg.hidden_size)

        if self.use_edge_readout:
            self.edge_projection = get_linear(
                in_features=global_cfg.hidden_size * (global_cfg.num_layers + 1),
                out_features=global_cfg.hidden_size,
                activation=Activation(global_cfg.activation),
                bias=True,
            )
            self.output_norm_edge = get_normalization_layer(
                NormalizationType(reg_cfg.normalization)
            )(global_cfg.hidden_size)

        if self.use_global_readout:
            self.global_projection = get_linear(
                in_features=global_cfg.hidden_size * (global_cfg.num_layers + 1),
                out_features=global_cfg.hidden_size,
                activation=Activation(global_cfg.activation),
                bias=True,
            )
            self.output_norm_global = get_normalization_layer(
                NormalizationType(reg_cfg.normalization)
            )(global_cfg.hidden_size)

    def forward(self, data, global_readouts, node_readouts, edge_readouts):
        node_features = self.node_projection(node_readouts)
        node_features = self.output_norm_node(node_features)

        if self.use_global_readout:
            global_features = self.global_projection(global_readouts)
            global_features = self.output_norm_global(global_features)
        else:
            global_features = None

        if self.use_edge_readout:
            edge_features = self.edge_projection(edge_readouts)
            edge_features = self.output_norm_edge(edge_features)
        else:
            edge_features = None
        return global_features, node_features, edge_features


class OutputLayer(nn.Module):
    """
    Get the final prediction from the readouts (force or energy)
    """

    def __init__(
        self,
        global_cfg: GlobalConfigs,
        gnn_cfg: GraphNeuralNetworksConfigs,
        reg_cfg: RegularizationConfigs,
        output_type: Literal["Vector", "Scalar"],
    ):
        super().__init__()

        self.output_type = output_type
        output_type_dict = {
            "Vector": 3,
            "Scalar": 1,
        }
        assert output_type in output_type_dict, f"Invalid output type {output_type}"

        # mlp
        self.ffn = get_feedforward(
            hidden_dim=global_cfg.hidden_size,
            activation=Activation(global_cfg.activation),
            hidden_layer_multiplier=gnn_cfg.output_hidden_layer_multiplier,
            dropout=reg_cfg.scalar_output_dropout
            if output_type == "Scalar"
            else reg_cfg.vector_output_dropout,
            bias=True,
        )

        # normalization
        # self.norm = get_normalization_layer(reg_cfg.normalization)(
        #     global_cfg.hidden_size
        # )

        # final output layer
        self.final_output = get_linear(
            in_features=global_cfg.hidden_size,
            out_features=output_type_dict[output_type],
            activation=None,
        )

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        """
        features: features from the backbone
        Shape ([num_nodes, hidden_size] or [num_nodes, max_neighbor, hidden_size])
        """
        # mlp
        features = self.ffn(features)

        # final output layer
        return self.final_output(features)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse
import logging
from pathlib import Path

from fairchem.experimental.legacy.utils import load_config

from fairchem.core.common.registry import registry
from fairchem.core.modules.normalization._load_utils import save_checkpoint
from fairchem.core.modules.normalization.element_references import fit_linear_references


def fit_linref(config: dict, output_path: str | Path) -> None:
    """Fit linear references using the standard config

    Args:
        config: config
        output_path: output path
    """
    # load the training dataset
    output_path = Path(output_path).resolve()

    try:
        # load the training dataset
        train_dataset = registry.get_dataset_class(
            config["dataset"]["train"].get("format", "lmdb")
        )(config["dataset"]["train"])
    except KeyError as err:
        raise ValueError("Train dataset is not specified in config!") from err

    try:
        elementref_config = config["dataset"]["train"]["transforms"][
            "element_references"
        ]["fit"]
    except KeyError as err:
        raise ValueError(
            "The provided config does not specify a 'fit' block for 'element_refereces'!"
        ) from err

    element_refs = fit_linear_references(
        targets=elementref_config["targets"],
        dataset=train_dataset,
        batch_size=elementref_config.get("batch_size", 32),
        num_batches=elementref_config.get("num_batches"),
        num_workers=config.get("optim", {}).get("num_workers", 16),
        max_num_elements=elementref_config.get("max_num_elements", 118),
        driver=elementref_config.get("driver", None),
    )

    for target, references in element_refs.items():
        path = save_checkpoint(
            references.state_dict(),
            output_path,
            f"{target}_linref.pt",
        )
        logging.info(f"{target} linear references have been saved to: {path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        required=True,
        type=Path,
        help="Path to configuration yaml file",
    )
    parser.add_argument(
        "--out-path",
        default=".",
        type=str,
        help="Output path to save linear references",
    )
    args = parser.parse_args()
    config, dup_warning, dup_error = load_config(args.config)

    if len(dup_warning) > 0:
        logging.warning(
            f"The following keys in the given config have duplicates: {dup_warning}."
        )
    if len(dup_error) > 0:
        raise RuntimeError(
            f"The following include entries in the config have duplicates: {dup_error}"
        )

    fit_linref(config, args.out_path)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ase.io.jsonio import encode

from fairchem.core.components.calculate.recipes.relax import (
    relax_atoms,
)

if TYPE_CHECKING:
    from ase import Atoms
    from ase.calculators.calculator import Calculator
    from ase.optimize.optimize import Optimizer


def adsorb_atoms(
    adslab_atoms: Atoms,
    calculator: Calculator,
    optimizer_cls: Optimizer,
    steps: int = 500,
    fmax: float = 0.02,
    relax_surface: bool = False,
    save_relaxed_atoms: bool = False,
) -> Atoms:
    """
    Simple helper function to run relaxations and compute the adsorption energy
    of a given adsorbate+surface atoms object.

    Args:
        atoms: ASE atoms object
        calculator: ASE calculator
        optimizer_cls: ASE optimizer. Default LBFGS
        steps: max number of relaxation steps
        fmax: force convergence threshold
        relax_surface: Whether to relax the bare surface
        save_relaxed_atoms: Whether to save the relaxed atoms
    Returns:
        dict of adsorption results
    """
    sid = adslab_atoms.info["sid"]
    gas_reference_energy = adslab_atoms.info["gas_ref"]
    # Compute DFT adsorption energy
    dft_relaxed_adslab_energy = adslab_atoms.info["dft_relaxed_adslab_energy"]
    dft_relaxed_slab_energy = adslab_atoms.info["dft_relaxed_slab_energy"]
    dft_adsorption_energy = (
        dft_relaxed_adslab_energy - dft_relaxed_slab_energy - gas_reference_energy
    )

    # Relax provided adslab system
    adslab_atoms.calc = calculator
    relaxed_adslab_atoms = relax_atoms(
        adslab_atoms,
        fmax=fmax,
        steps=steps,
        optimizer_cls=optimizer_cls,
    )
    relaxed_adslab_atoms_energy = relaxed_adslab_atoms.get_potential_energy()

    results = {
        "sid": sid,
        "direct": relaxed_adslab_atoms_energy,
        "target": dft_adsorption_energy,
    }

    # Compute adsorption energy using DFT slab energies
    hybrid_adsorption_energy = (
        relaxed_adslab_atoms_energy - dft_relaxed_slab_energy - gas_reference_energy
    )
    results["hybrid"] = hybrid_adsorption_energy

    # Compute adsorption energy using ML relaxed slabs
    if relax_surface:
        slab_atoms = adslab_atoms.info["initial_slab_atoms"]
        slab_atoms.calc = calculator
        relaxed_slab_atoms = relax_atoms(
            slab_atoms,
            fmax=fmax,
            steps=steps,
            optimizer_cls=optimizer_cls,
        )
        pred_adsorption_energy = (
            relaxed_adslab_atoms_energy
            - relaxed_slab_atoms.get_potential_energy()
            - gas_reference_energy
        )
        results["full"] = pred_adsorption_energy

    if save_relaxed_atoms:
        results["relaxed_adslab_atoms"] = encode(relaxed_adslab_atoms)
        if relax_surface:
            results["relaxed_slab_atoms"] = encode(relaxed_slab_atoms)

    return results



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse

from fairchem.core.models.equiformer_v2.eqv2_to_eqv2_hydra import (
    convert_checkpoint_and_config_to_hydra,
)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--eqv2-checkpoint", help="path to eqv2 checkpoint", type=str, required=True
    )
    parser.add_argument(
        "--eqv2-yaml", help="path to eqv2 yaml config", type=str, required=True
    )
    parser.add_argument(
        "--hydra-eqv2-checkpoint",
        help="path where to output hydra checkpoint",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--hydra-eqv2-yaml",
        help="path where to output hydra yaml",
        type=str,
        required=True,
    )
    args = parser.parse_args()

    convert_checkpoint_and_config_to_hydra(
        yaml_fn=args.eqv2_yaml,
        checkpoint_fn=args.eqv2_checkpoint,
        new_yaml_fn=args.hydra_eqv2_yaml,
        new_checkpoint_fn=args.hydra_eqv2_checkpoint,
    )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import copy
import logging
import math
import os
import random
import sys
from collections import defaultdict
from contextlib import nullcontext
from functools import wraps
from typing import TYPE_CHECKING, Protocol, Sequence

import hydra
import numpy as np
import torch
import torch.distributed as dist
from monty.dev import requires
from torch.distributed.elastic.utils.distributed import get_free_port
from torchtnt.framework import PredictUnit, State

from fairchem.core.common import gp_utils
from fairchem.core.common.distutils import (
    CURRENT_DEVICE_TYPE_STR,
    assign_device_for_local_rank,
    get_device_for_local_rank,
    setup_env_local_multi_gpu,
)
from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.units.mlip_unit import InferenceSettings
from fairchem.core.units.mlip_unit.utils import (
    load_inference_model,
    tf32_context_manager,
)

if TYPE_CHECKING:
    from fairchem.core.units.mlip_unit.mlip_unit import Task

try:
    import ray
    from ray import remote
    from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy

    ray_installed = True
except ImportError:
    ray = None

    def remote(cls):
        # dummy
        return cls

    ray_installed = False


def collate_predictions(predict_fn):
    @wraps(predict_fn)
    def collated_predict(
        predict_unit, data: AtomicData, undo_element_references: bool = True
    ):
        # Get the full prediction dictionary from the original predict method
        preds = predict_fn(predict_unit, data, undo_element_references)
        collated_preds = defaultdict(list)
        for i, dataset in enumerate(data.dataset):
            for task in predict_unit.dataset_to_tasks[dataset]:
                if task.level == "system":
                    collated_preds[task.property].append(
                        preds[task.name][i].unsqueeze(0)
                    )
                elif task.level == "atom":
                    collated_preds[task.property].append(
                        preds[task.name][data.batch == i]
                    )
                else:
                    raise RuntimeError(
                        f"Unrecognized task level={task.level} found in data batch at position {i}"
                    )

        return {prop: torch.cat(val) for prop, val in collated_preds.items()}

    return collated_predict


class MLIPPredictUnitProtocol(Protocol):
    def predict(self, data: AtomicData, undo_element_references: bool) -> dict: ...

    @property
    def dataset_to_tasks(self) -> dict[str, list]: ...


def merge_uma_model(model, data):
    # merge the backbone
    model.backbone = model.backbone.merge_MOLE_model(data)

    # merge any heads
    new_output_heads = torch.nn.ModuleDict()
    for head_name, head in model.output_heads.items():
        if hasattr(head, "merge_MOLE_model"):
            new_output_heads[head_name] = head.merge_MOLE_model(data)
        else:
            new_output_heads[head_name] = head
    model.output_heads = new_output_heads


class MLIPPredictUnit(PredictUnit[AtomicData], MLIPPredictUnitProtocol):
    def __init__(
        self,
        inference_model_path: str,
        device: str = "cpu",
        overrides: dict | None = None,
        inference_settings: InferenceSettings | None = None,
        seed: int = 41,
        atom_refs: dict | None = None,
        assert_on_nans: bool = False,
    ):
        super().__init__()
        os.environ[CURRENT_DEVICE_TYPE_STR] = device

        self.set_seed(seed)
        # note these are different from the element references used for model training
        self.atom_refs = (
            {task.replace("_elem_refs", ""): refs for task, refs in atom_refs.items()}
            if atom_refs is not None
            else {}
        )

        if inference_settings is None:
            inference_settings = InferenceSettings()
        if inference_settings.torch_num_threads is not None:
            torch.set_num_threads(inference_settings.torch_num_threads)
            torch.set_num_interop_threads(inference_settings.torch_num_threads)

        if overrides is None:
            overrides = {}
        if "backbone" not in overrides:
            overrides["backbone"] = {}
        # always disable always_use_pbc for inference
        overrides["backbone"]["always_use_pbc"] = False
        if inference_settings.activation_checkpointing is not None:
            overrides["backbone"]["activation_checkpointing"] = (
                inference_settings.activation_checkpointing
            )
        if inference_settings.edge_chunk_size is not None:
            overrides["backbone"]["edge_chunk_size"] = (
                inference_settings.edge_chunk_size
            )
        if inference_settings.external_graph_gen is not None:
            overrides["backbone"][
                "otf_graph"
            ] = not inference_settings.external_graph_gen

        if inference_settings.internal_graph_gen_version is not None:
            overrides["backbone"]["radius_pbc_version"] = (
                inference_settings.internal_graph_gen_version
            )

        if inference_settings.wigner_cuda:
            logging.warning(
                "The wigner_cuda flag is deprecated and will be removed in future versions."
            )

        self.model, checkpoint = load_inference_model(
            inference_model_path, use_ema=True, overrides=overrides
        )
        tasks = [
            hydra.utils.instantiate(task_config)
            for task_config in checkpoint.tasks_config
        ]
        self.tasks = {t.name: t for t in tasks}

        self._dataset_to_tasks = get_dataset_to_tasks_map(self.tasks.values())
        assert set(self._dataset_to_tasks.keys()).issubset(
            set(self.model.module.backbone.dataset_list)
        ), "Datasets in tasks is not a strict subset of datasets in backbone."
        assert device in ["cpu", "cuda"], "device must be either 'cpu' or 'cuda'"

        self.device = get_device_for_local_rank() if device == "cuda" else "cpu"

        self.model.eval()

        self.lazy_model_intialized = False
        self.inference_settings = inference_settings

        # store composition embedding of system the model was merged on
        self.merged_on = None

        self.assert_on_nans = assert_on_nans

        if self.direct_forces:
            logging.warning(
                "This is a direct-force model. Direct force predictions may lead to discontinuities in the potential "
                "energy surface and energy conservation errors."
            )

    @property
    def direct_forces(self) -> bool:
        return self.model.module.backbone.direct_forces

    @property
    def dataset_to_tasks(self) -> dict[str, list]:
        return self._dataset_to_tasks

    def set_seed(self, seed: int):
        logging.debug(f"Setting random seed to {seed}")
        self._seed = seed
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    def move_to_device(self):
        self.model.to(self.device)
        for task in self.tasks.values():
            task.normalizer.to(self.device)
            if task.element_references is not None:
                task.element_references.to(self.device)

    def predict_step(self, state: State, data: AtomicData) -> dict[str, torch.tensor]:
        return self.predict(data)

    def get_composition_charge_spin_dataset(self, data):
        composition_sum = data.atomic_numbers.new_zeros(
            self.model.module.backbone.max_num_elements,
            dtype=torch.int,
        ).index_add(
            0,
            data.atomic_numbers.to(torch.int),
            data.atomic_numbers.new_ones(data.atomic_numbers.shape[0], dtype=torch.int),
        )
        comp_charge_spin = (
            composition_sum,
            getattr(data, "charge", None),
            getattr(data, "spin", None),
        )
        return comp_charge_spin, getattr(data, "dataset", [None])

    @collate_predictions
    def predict(
        self, data: AtomicData, undo_element_references: bool = True
    ) -> dict[str, torch.tensor]:
        if not self.lazy_model_intialized:
            # merge everything on CPU
            if self.inference_settings.merge_mole:
                # replace backbone with non MOE version
                assert (
                    data.natoms.numel() == 1
                ), f"Cannot merge model with multiple systems in batch. Must be exactly 1 system, found {data.natoms.numel()}"

                merge_uma_model(self.model.module, data.clone())

                self.model.eval()
            # move to device
            self.move_to_device()
            if self.inference_settings.compile:
                logging.warning(
                    "Model is being compiled this might take a while for the first time"
                )
                self.model = torch.compile(self.model, dynamic=True)
            self.lazy_model_intialized = True

        # this needs to be .clone() to avoid issues with graph parallel modifying this data with MOLE
        data_device = data.to(self.device).clone()

        if self.inference_settings.merge_mole:
            if self.merged_on is None:
                # only get embeddings after moved to final device to get right types
                self.merged_on = self.get_composition_charge_spin_dataset(data_device)
            else:
                this_sys = self.get_composition_charge_spin_dataset(data_device)
                assert (
                    data_device.natoms.numel() == 1
                ), f"Cannot run merged model on batch with multiple systems. Must be exactly 1 system, found {data_device.natoms.numel()}"

                # Normalize compositions by total number of atoms to allow same reduced composition
                merged_comp = self.merged_on[0][0].float()
                this_comp = this_sys[0][0].float()
                merged_comp_norm = merged_comp / merged_comp.sum()
                this_comp_norm = this_comp / this_comp.sum()

                assert merged_comp_norm.isclose(
                    this_comp_norm, rtol=1e-5
                ).all(), "Cannot run on merged model on system. Relative compositions seem different..."
                assert (
                    self.merged_on[0][1] == this_sys[0][1]
                ), f"Cannot run on merged model on system. Charge is different {self.merged_on[0][1]} vs {this_sys[0][1]}"
                assert (
                    self.merged_on[0][2] == this_sys[0][2]
                ), f"Cannot run on merged model on system. Spin is different {self.merged_on[0][2]} vs {this_sys[0][2]}"
                assert (
                    self.merged_on[1] == this_sys[1]
                ), f"Cannot run on merged model on system. Dataset is different {self.merged_on[1]} vs {this_sys[1]}"

        inference_context = torch.no_grad() if self.direct_forces else nullcontext()
        tf32_context = (
            tf32_context_manager() if self.inference_settings.tf32 else nullcontext()
        )

        pred_output = {}
        with inference_context, tf32_context:
            output = self.model(data_device)
            for task_name, task in self.tasks.items():
                pred_output[task_name] = task.normalizer.denorm(
                    output[task_name][task.property]
                )
                if self.assert_on_nans:
                    assert torch.isfinite(
                        pred_output[task_name]
                    ).all(), f"NaNs/Infs found in prediction for task {task_name}.{task.property}"
                if undo_element_references and task.element_references is not None:
                    pred_output[task_name] = task.element_references.undo_refs(
                        data_device, pred_output[task_name]
                    )

        return pred_output


def get_dataset_to_tasks_map(tasks: Sequence[Task]) -> dict[str, list[Task]]:
    """Create a mapping from dataset names to their associated tasks.

    Args:
        tasks: A sequence of Task objects to be organized by dataset

    Returns:
        A dictionary mapping dataset names (str) to lists of Task objects
        that are associated with that dataset
    """
    dset_to_tasks_map = defaultdict(list)
    for task in tasks:
        for dataset_name in task.datasets:
            dset_to_tasks_map[dataset_name].append(task)
    return dict(dset_to_tasks_map)


def move_tensors_to_cpu(data):
    """
    Recursively move all PyTorch tensors in a nested data structure to CPU.

    Args:
        data: Input data structure (dict, list, tuple, tensor, or other)

    Returns:
        Data structure with all tensors moved to CPU
    """
    if isinstance(data, torch.Tensor):
        return data.cpu()
    elif isinstance(data, dict):
        return {key: move_tensors_to_cpu(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [move_tensors_to_cpu(item) for item in data]
    elif isinstance(data, tuple):
        return tuple(move_tensors_to_cpu(item) for item in data)
    else:
        # Return as-is for non-tensor types (str, int, float, etc.)
        return data


class MLIPWorkerLocal:
    def __init__(
        self,
        worker_id: int,
        world_size: int,
        predictor_config: dict,
        master_port: int | None = None,
        master_address: str | None = None,
    ):
        if ray_installed is False:
            raise RuntimeError("Requires `ray` to be installed")

        self.worker_id = worker_id
        self.world_size = world_size
        self.predictor_config = predictor_config
        self.master_address = (
            ray.util.get_node_ip_address() if master_address is None else master_address
        )
        self.master_port = get_free_port() if master_port is None else master_port
        self.is_setup = False
        self.last_received_atomic_data = None

    def get_master_address_and_port(self):
        return (self.master_address, self.master_port)

    def get_device_for_local_rank(self):
        return get_device_for_local_rank()

    def _distributed_setup(
        self,
    ):
        # initialize distributed environment
        # TODO, this wont work for multi-node, need to fix master addr
        logging.info(f"Initializing worker {self.worker_id}...")
        setup_env_local_multi_gpu(self.worker_id, self.master_port, self.master_address)

        device = self.predictor_config.get("device", "cpu")
        assign_device_for_local_rank(device == "cpu", 0)
        backend = "gloo" if device == "cpu" else "nccl"
        dist.init_process_group(
            backend=backend,
            rank=self.worker_id,
            world_size=self.world_size,
        )
        gp_utils.setup_graph_parallel_groups(self.world_size, backend)
        self.predict_unit = hydra.utils.instantiate(self.predictor_config)
        self.device = get_device_for_local_rank()
        logging.info(
            f"Worker {self.worker_id}, gpu_id: {ray.get_gpu_ids()}, loaded predict unit: {self.predict_unit}, "
            f"on port {self.master_port}, with device: {self.device}, config: {self.predictor_config}"
        )
        self.is_setup = True

    def predict(
        self, data: AtomicData, use_nccl: bool = False
    ) -> dict[str, torch.tensor] | None:
        if not self.is_setup:
            self._distributed_setup()

        out = self.predict_unit.predict(data)
        if self.worker_id == 0:
            return move_tensors_to_cpu(out)

        if self.worker_id != 0 and use_nccl:
            self.last_received_atomic_data = data.to(self.device)
            while True:
                torch.distributed.broadcast(self.last_received_atomic_data.pos, src=0)
                self.predict_unit.predict(self.last_received_atomic_data)

        return None


@remote
class MLIPWorker(MLIPWorkerLocal):
    pass


@requires(ray_installed, message="Requires `ray` to be installed")
class ParallelMLIPPredictUnit(MLIPPredictUnitProtocol):
    def __init__(
        self,
        inference_model_path: str,
        device: str = "cpu",
        overrides: dict | None = None,
        inference_settings: InferenceSettings | None = None,
        seed: int = 41,
        atom_refs: dict | None = None,
        assert_on_nans: bool = False,
        num_workers: int = 1,
        num_workers_per_node: int = 8,
        log_level: int = logging.INFO,
    ):
        super().__init__()
        _mlip_pred_unit = MLIPPredictUnit(
            inference_model_path=inference_model_path,
            device="cpu",
            overrides=overrides,
            inference_settings=inference_settings,
            seed=seed,
            atom_refs=atom_refs,
        )
        self.inference_settings = inference_settings
        self._dataset_to_tasks = copy.deepcopy(_mlip_pred_unit.dataset_to_tasks)

        predict_unit_config = {
            "_target_": "fairchem.core.units.mlip_unit.predict.MLIPPredictUnit",
            "inference_model_path": inference_model_path,
            "device": device,
            "overrides": overrides,
            "inference_settings": inference_settings,
            "seed": seed,
            "atom_refs": atom_refs,
            "assert_on_nans": assert_on_nans,
        }

        logging.basicConfig(
            level=log_level,
            force=True,
            stream=sys.stdout,
            format="%(asctime)s %(levelname)s [%(processName)s] %(name)s: %(message)s",
        )
        # Optional: keep Ray/uvicorn chatty logs in check
        logging.getLogger("ray").setLevel(log_level)
        logging.getLogger("uvicorn").setLevel(log_level)
        if not ray.is_initialized():
            # in CI envrionment, we want to control the number of CPUs allocated to limit the pool of IDLE ray workers
            if os.environ.get("CI"):
                logging.info(
                    f"CI environment detected, initializing ray with limited CPUs: {num_workers_per_node}"
                )
                ray.init(
                    logging_level=log_level,
                    num_cpus=num_workers_per_node,
                    runtime_env={
                        "env_vars": {"RAY_DEBUG": "1"},
                    },
                )
            else:
                ray.init(logging_level=log_level)

        self.atomic_data_on_device = None

        num_nodes = math.ceil(num_workers / num_workers_per_node)
        num_workers_on_node_array = [num_workers_per_node] * num_nodes
        if num_workers % num_workers_per_node > 0:
            num_workers_on_node_array[-1] = num_workers % num_workers_per_node
        logging.info(
            f"Creating placement groups with {num_workers_on_node_array} workers on {device}"
        )

        # first create one placement group for each node
        num_gpu_per_worker = 1 if device == "cuda" else 0
        placement_groups = []
        for workers in num_workers_on_node_array:
            bundle = {"CPU": workers}
            if device == "cuda":
                bundle["GPU"] = workers
            pg = ray.util.placement_group([bundle], strategy="STRICT_PACK")
            placement_groups.append(pg)
        ray.get(pg.ready())  # Wait for each placement group to be scheduled

        # Need to still place worker to occupy space, otherwise ray double books this GPU
        rank0_worker = MLIPWorker.options(
            num_gpus=num_gpu_per_worker,
            scheduling_strategy=PlacementGroupSchedulingStrategy(
                placement_group=placement_groups[0],
                placement_group_bundle_index=0,  # Use the first (and only) bundle in the PG
                placement_group_capture_child_tasks=True,  # Ensure child tasks also run in this PG
            ),
        ).remote(0, num_workers, predict_unit_config)

        local_gpu_or_cpu = ray.get(rank0_worker.get_device_for_local_rank.remote())
        os.environ[CURRENT_DEVICE_TYPE_STR] = local_gpu_or_cpu

        self.workers = []
        self.local_rank0 = MLIPWorkerLocal(
            worker_id=0,
            world_size=num_workers,
            predictor_config=predict_unit_config,
        )
        master_addr, master_port = self.local_rank0.get_master_address_and_port()
        logging.info(f"Started rank0 on {master_addr}:{master_port}")

        # next place all ranks in order and pack them on placement groups
        # ie: rank0-7 -> placement group 0, 8->15 -> placement group 1 etc.
        worker_id = 0
        for pg_idx, pg in enumerate(placement_groups):
            workers = num_workers_on_node_array[pg_idx]
            logging.info(
                f"Launching workers for placement group {pg_idx} (Node {pg_idx}), workers={workers}"
            )

            for i in range(workers):
                # skip the first one because it's already been initialized above
                if pg_idx == 0 and i == 0:
                    worker_id += 1
                    continue
                # Each actor requests 1 worker worth of resources and uses the specific placement group
                actor = MLIPWorker.options(
                    num_gpus=num_gpu_per_worker,
                    scheduling_strategy=PlacementGroupSchedulingStrategy(
                        placement_group=pg,
                        placement_group_bundle_index=0,  # Use the first (and only) bundle in the PG
                        placement_group_capture_child_tasks=True,  # Ensure child tasks also run in this PG
                    ),
                ).remote(
                    worker_id,
                    num_workers,
                    predict_unit_config,
                    master_port,
                    master_addr,
                )
                self.workers.append(actor)
                worker_id += 1

    def predict(self, data: AtomicData) -> dict[str, torch.tensor]:
        # put the reference in the object store only once
        if not self.inference_settings.merge_mole or self.atomic_data_on_device is None:
            data_ref = ray.put(data)
            # this will put the ray works into an infinite loop listening for broadcasts
            _futures = [
                w.predict.remote(data_ref, use_nccl=self.inference_settings.merge_mole)
                for w in self.workers
            ]
            self.atomic_data_on_device = data.clone()
        else:
            self.atomic_data_on_device.pos = data.pos.to(self.local_rank0.device)
            torch.distributed.broadcast(self.atomic_data_on_device.pos, src=0)

        return self.local_rank0.predict(self.atomic_data_on_device)

    @property
    def dataset_to_tasks(self) -> dict[str, list]:
        return self._dataset_to_tasks



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging

import ase.units
import numpy as np
import pandas as pd
from ase.io.jsonio import decode
from monty.dev import requires
from tqdm import tqdm

from fairchem.core.components.benchmark import JsonDFReducer
from fairchem.core.components.calculate import RelaxationRunner, SinglePointRunner
from fairchem.core.components.calculate.recipes.local_env import construct_bond_matrix

try:
    from scipy.stats import kendalltau, spearmanr
    from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score

    sklearn_scipy_installed = True
except ImportError:
    sklearn_scipy_installed = False

try:
    from pymatgen.analysis.local_env import JmolNN
    from pymatgen.analysis.structure_matcher import StructureMatcher
    from pymatgen.io.ase import AseAtomsAdaptor

    pmg_installed = True
except ImportError:
    pmg_installed = False

ev2kJ = ase.units.eV * ase.units.mol / ase.units.kJ


@requires(pmg_installed, message="Requires `pymatgen` to be installed")
@requires(
    sklearn_scipy_installed, "Requires `scipy` and `scikit-learn` to be installed"
)
class OMCPolymorphReducer(JsonDFReducer):
    def __init__(
        self,
        benchmark_name: str,
        target_data_key: str,
        molecule_id_key: str,
        calculate_structural_metrics: bool = False,
        index_name: str | None = None,
    ):
        """
        Args:
            benchmark_name: Name of the benchmark, used for file naming
            target_data_key: Name of the normalized target energy
            molecule_id_key: Key name of molecule to identify polymorphs
            calculate_structural_metrics: Whether to calculate structural metrics: match rate, and root mean squared
                distance of atomic positions for matched structures.
                A structure is considered a match if atomic positions match within given tolerance and no
                and matrices of covelent bond networks computed using JmolNN between target and relaxed structure
                are identical. RMSD is only computed if a structures is a match.
            index_name: Optional name of the column to use as index
        """
        self._molecule_id_key = molecule_id_key
        self._calc_structural_metrics = calculate_structural_metrics

        if self._calc_structural_metrics:
            self._structure_matcher = StructureMatcher(
                primitive_cell=False
            )  # ltol=0.5, stol=0.5, angle_tol=10,
            self._jmolnn = JmolNN()
        else:
            self._structure_matcher = None
            self._jmolnn = None

        super().__init__(
            benchmark_name=benchmark_name,
            index_name=index_name,
            target_data_keys=[target_data_key],
        )

    @property
    def runner_type(self) -> type[SinglePointRunner | RelaxationRunner]:
        """The runner type this reducer is associated with."""
        return RelaxationRunner if self._calc_structural_metrics else SinglePointRunner

    def compute_metrics(self, results: pd.DataFrame, run_name: str) -> pd.DataFrame:
        """Compute OMC polymorph metrics for single point or relaxed energy and structure predictions.

        Args:
            results: DataFrame containing prediction results with energy values
            run_name: Identifier for the current evaluation run

        Returns:
            DataFrame containing computed metrics for different material subsets
        """
        metrics = {}
        energy_key = self.target_data_keys[0]
        for molecule_id in results[self._molecule_id_key].unique():
            logging.info(f"Computing metrics for {molecule_id=}")
            polymorph_results = results[results[self._molecule_id_key] == molecule_id]

            ref_structure_id = polymorph_results[f"{energy_key}_target"].argmin()
            ref_energy = polymorph_results[energy_key].iloc[ref_structure_id]
            ref_energy_target = polymorph_results[f"{energy_key}_target"].iloc[
                ref_structure_id
            ]

            energy_latt = polymorph_results[energy_key] - ref_energy
            energy_latt_target = (
                polymorph_results[f"{energy_key}_target"] - ref_energy_target
            )

            rankings = pd.Series(energy_latt).rank(method="min")
            rankings_target = pd.Series(energy_latt_target).rank(method="min")

            metrics.update(
                {
                    f"{molecule_id},mae": ev2kJ
                    * mean_absolute_error(energy_latt_target, energy_latt),
                    f"{molecule_id},rmse": ev2kJ
                    * np.sqrt(mean_squared_error(energy_latt_target, energy_latt)),
                    f"{molecule_id},r2": r2_score(energy_latt_target, energy_latt),
                    f"{molecule_id},kendall": kendalltau(
                        rankings_target, rankings
                    ).statistic,
                    f"{molecule_id},spearmanr": spearmanr(
                        rankings_target, rankings
                    ).statistic,
                }
            )

            if self._calc_structural_metrics:  # TODO this will need to be parallelized
                _rmsds = []
                for index in tqdm(
                    polymorph_results.index,
                    desc=f"Matching polymorphs for {molecule_id=}",
                ):
                    entry = polymorph_results.loc[index]
                    relaxed_structure = AseAtomsAdaptor.get_structure(
                        decode(entry["atoms"])
                    )
                    reference_structure = AseAtomsAdaptor.get_structure(
                        decode(entry["atoms_relaxed_target"])
                    )

                    # not clean but call this directly to avoid rematching (rms, max_dist, mask, cost, mapping)
                    reference_structure, relaxed_structure, _, _ = (
                        self._structure_matcher._preprocess(
                            reference_structure, relaxed_structure, niggli=True
                        )
                    )
                    match = self._structure_matcher._match(
                        reference_structure, relaxed_structure, fu=1, use_rms=True
                    )
                    if match is not None:
                        reference_matrix = construct_bond_matrix(
                            reference_structure, nn_finder=self._jmolnn
                        )
                        relaxed_matrix = construct_bond_matrix(
                            relaxed_structure,
                            nn_finder=self._jmolnn,
                            site_permutations=match[4],
                        )
                        if np.array_equal(relaxed_matrix, reference_matrix) is True:
                            # rmsd from pmg structure matcher is normalized by (V/nsites)^(1/3)
                            avg_vol = (
                                relaxed_structure.volume + reference_structure.volume
                            ) / 2
                            rmsd = match[0] * (avg_vol / len(reference_structure)) ** (
                                1 / 3
                            )
                            _rmsds.append(rmsd)

                metrics.update(
                    {
                        f"{molecule_id},rmsd": np.mean(_rmsds) if _rmsds else 0.0,
                        f"{molecule_id},rmsd_std": np.std(_rmsds) if _rmsds else 0.0,
                        f"{molecule_id},match_rate": len(_rmsds)
                        / len(polymorph_results),
                    }
                )

        return pd.DataFrame([metrics], index=[run_name])

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True

    def load_state(self, checkpoint_location: str | None) -> None:
        pass



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

import torch
import wandb

from fairchem.core.common import distutils
from fairchem.core.common.registry import registry
from fairchem.core.common.utils import tensor_stats

if TYPE_CHECKING:
    from pandas import DataFrame


class Logger(ABC):
    """Generic class to interface with various logging modules, e.g. wandb,
    tensorboard, etc.
    """

    def __init__(self, config) -> None:
        self.config = config

    @abstractmethod
    def watch(self, model, log_freq: int = 1000):
        """
        Monitor parameters and gradients.
        """

    def log(self, update_dict, step: int, split: str = ""):
        """
        Log some values.
        """
        assert step is not None
        if split != "":
            new_dict = {}
            for key in update_dict:
                new_dict[f"{split}/{key}"] = update_dict[key]
            update_dict = new_dict
        return update_dict

    @abstractmethod
    def log_plots(self, plots) -> None:
        pass

    @abstractmethod
    def mark_preempting(self) -> None:
        pass

    @abstractmethod
    def log_summary(self, summary_dict: dict[str, Any]) -> None:
        pass

    @abstractmethod
    def log_artifact(self, name: str, type: str, file_location: str) -> None:
        pass


@registry.register_logger("wandb")
class WandBLogger(Logger):
    def __init__(self, config) -> None:
        super().__init__(config)
        project = (
            self.config["logger"].get("project", None)
            if isinstance(self.config["logger"], dict)
            else None
        )
        entity = (
            self.config["logger"].get("entity", None)
            if isinstance(self.config["logger"], dict)
            else None
        )
        group = (
            self.config["logger"].get("group", None)
            if isinstance(self.config["logger"], dict)
            else None
        )

        wandb.init(
            config=self.config,
            id=self.config["cmd"]["timestamp_id"],
            name=self.config["cmd"]["identifier"],
            dir=self.config["cmd"]["logs_dir"],
            project=project,
            entity=entity,
            resume="allow",
            group=group,
        )

    def watch(self, model, log="all", log_freq: int = 1000) -> None:
        wandb.watch(model, log=log, log_freq=log_freq)

    def log(self, update_dict, step: int, split: str = "") -> None:
        update_dict = super().log(update_dict, step, split)
        wandb.log(update_dict, step=int(step))

    def log_plots(self, plots, caption: str = "") -> None:
        assert isinstance(plots, list)
        plots = [wandb.Image(x, caption=caption) for x in plots]
        wandb.log({"data": plots})

    def log_table(
        self,
        name: str,
        cols: list,
        data: list,
        step: int | None = None,
        commit: bool = False,
    ) -> None:
        # cols are 1D list of N elements, data must be NxK where the number of cols must match cols
        # see https://docs.wandb.ai/guides/tables
        table = wandb.Table(columns=cols, data=data)
        wandb.log({name: table}, step=step, commit=commit)

    def log_dataframe(
        self,
        name: str,
        dataframe: DataFrame,
        step: int | None = None,
        commit: bool = False,
    ):
        table = wandb.Table(dataframe=dataframe)
        wandb.log({name: table}, step=step, commit=commit)

    def log_summary(self, summary_dict: dict[str, Any]):
        for k, v in summary_dict.items():
            wandb.run.summary[k] = v

    def mark_preempting(self) -> None:
        wandb.mark_preempting()

    def log_artifact(self, name: str, type: str, file_location: str) -> None:
        art = wandb.Artifact(name=name, type=type)
        art.add_file(file_location)
        art.save()


@registry.register_logger("tensorboard")
class TensorboardLogger(Logger):
    def __init__(self, config) -> None:
        try:
            from torch.utils.tensorboard import SummaryWriter
        except ImportError as err:
            raise err
        finally:
            warnings.warn("Tensorboard logger is deprecated.", DeprecationWarning())

        super().__init__(config)
        self.writer = SummaryWriter(self.config["cmd"]["logs_dir"])

    # TODO: add a model hook for watching gradients.
    def watch(self, model, log_freq: int = 1000) -> bool:
        logging.warning("Model gradient logging to tensorboard not yet supported.")
        return False

    def log(self, update_dict, step: int, split: str = ""):
        update_dict = super().log(update_dict, step, split)
        for key in update_dict:
            if torch.is_tensor(update_dict[key]):
                self.writer.add_scalar(key, update_dict[key].item(), step)
            else:
                assert isinstance(update_dict[key], (int, float))
                self.writer.add_scalar(key, update_dict[key], step)

    def mark_preempting(self) -> None:
        logging.warning("mark_preempting for Tensorboard not supported")

    def log_plots(self, plots) -> None:
        logging.warning("log_plots for Tensorboard not supported")

    def log_summary(self, summary_dict: dict[str, Any]) -> None:
        logging.warning("log_summary for Tensorboard not supported")

    def log_artifact(self, name: str, type: str, file_location: str) -> None:
        logging.warning("log_artifact for Tensorboard not supported")


class WandBSingletonLogger:
    """
    Singleton version of wandb logger, this forces a single instance of the logger to be created and used from anywhere in the code (not just from the trainer).
    This will replace the original WandBLogger.

    We initialize wandb instance somewhere in the trainer/runner globally:

    WandBSingletonLogger.init_wandb(...)

    Then from anywhere in the code we can fetch the singleton instance and log to wandb,
    note this allows you to log without knowing explicitly which step you are on
    see: https://docs.wandb.ai/ref/python/log/#the-wb-step for more details

    WandBSingletonLogger.get_instance().log({"some_value": value}, commit=False)
    """

    _instance = None

    def __init__(self):
        raise RuntimeError("Call get_instance() instead")

    @classmethod
    def initialized(cls) -> bool:
        return WandBSingletonLogger._instance is not None

    @classmethod
    def init_wandb(
        cls,
        config: dict,
        run_id: str,
        run_name: str,
        log_dir: str,
        project: str,
        entity: str,
        group: str | None = None,
        job_type: str | None = None,
    ) -> None:
        wandb.init(
            config=config,
            id=run_id,
            name=run_name,
            dir=log_dir,
            project=project,
            entity=entity,
            resume="allow",
            group=group,
            job_type=job_type,
        )

    @classmethod
    def get_instance(cls):
        assert wandb.run is not None, "wandb is not initialized, call init_wandb first!"
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
        return cls._instance

    def watch(self, model, log="all", log_freq: int = 1000) -> None:
        wandb.watch(model, log=log, log_freq=log_freq)

    def log(
        self, update_dict: dict, step: int | None = None, commit=False, split: str = ""
    ) -> None:
        # HACK: this is really ugly logic here for backward compat but we should get rid of.
        # the split string shouldn't inserted here
        if split != "":
            new_dict = {}
            for key in update_dict:
                new_dict[f"{split}/{key}"] = update_dict[key]
            update_dict = new_dict

        # if step is not specified, wandb will use an auto-incremented step: https://docs.wandb.ai/ref/python/log/
        # otherwise the user must increment it manually (not recommended)
        wandb.log(update_dict, step=step, commit=commit)

    def log_table(
        self, name: str, cols: list, data: list, step: int | None = None, commit=False
    ) -> None:
        # cols are 1D list of N elements, data must be NxK where the number of cols must match cols
        # see https://docs.wandb.ai/guides/tables
        table = wandb.Table(columns=cols, data=data)
        wandb.log({name: table}, step=step, commit=commit)

    def log_dataframe(
        self,
        name: str,
        dataframe: DataFrame,
        step: int | None = None,
        commit: bool = False,
    ):
        table = wandb.Table(dataframe=dataframe)
        wandb.log({name: table}, step=step, commit=commit)

    def log_summary(self, summary_dict: dict[str, Any]):
        for k, v in summary_dict.items():
            wandb.run.summary[k] = v

    def mark_preempting(self) -> None:
        wandb.mark_preempting()

    def log_artifact(self, name: str, type: str, file_location: str) -> None:
        art = wandb.Artifact(name=name, type=type)
        art.add_file(file_location)
        art.save()


# convienience function for logging stats with WandBSingletonLogger
def log_stats(x: torch.Tensor, prefix: str):
    if distutils.is_master() and WandBSingletonLogger._instance is not None:
        WandBSingletonLogger.get_instance().log(
            tensor_stats(prefix, x),
            commit=False,
        )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import functools
import logging
import os
import random
import timeit
import uuid
from collections import defaultdict

import numpy as np
import torch
from ase.build import make_supercell
from ase.io import read
from torch.profiler import ProfilerActivity, profile

from fairchem.core.common.profiler_utils import get_profile_schedule
from fairchem.core.components.runner import Runner
from fairchem.core.datasets.atomic_data import AtomicData, atomicdata_list_to_batch
from fairchem.core.datasets.common_structures import get_fcc_carbon_xtal
from fairchem.core.units.mlip_unit import MLIPPredictUnit
from fairchem.core.units.mlip_unit.api.inference import (
    InferenceSettings,
    inference_settings_default,
)


def seed_everywhere(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def ase_to_graph(
    atoms, neighbors: int, cutoff: float, external_graph=True, dataset_name="omat"
):
    data_object = AtomicData.from_ase(
        atoms,
        max_neigh=neighbors,
        radius=cutoff,
        r_edges=external_graph,
        task_name=dataset_name,
    )
    data_object.natoms = torch.tensor(len(atoms))
    data_object.charge = torch.LongTensor([0])
    data_object.spin = torch.LongTensor([0])
    data_object.pos.requires_grad = True
    return atomicdata_list_to_batch([data_object])


def get_qps(data, predictor, warmups: int = 10, timeiters: int = 100):
    def timefunc():
        predictor.predict(data)
        torch.cuda.synchronize()

    for _ in range(warmups):
        timefunc()
        logging.info(f"memory allocated: {torch.cuda.memory_allocated()/(1024**3)}")

    result = timeit.timeit(timefunc, number=timeiters)
    qps = timeiters / result
    ns_per_day = qps * 24 * 3600 / 1e6
    return qps, ns_per_day


def trace_handler(p, name, save_loc):
    trace_name = f"{name}.pt.trace.json"
    output_path = os.path.join(save_loc, trace_name)
    logging.info(f"Saving trace in {output_path}")
    p.export_chrome_trace(output_path)


def make_profile(data, predictor, name, save_loc):
    activities = [ProfilerActivity.CPU, ProfilerActivity.CUDA]
    profile_schedule, total_profile_steps = get_profile_schedule()
    tc = functools.partial(trace_handler, name=name, save_loc=save_loc)

    with profile(
        activities=activities,
        schedule=profile_schedule,
        on_trace_ready=tc,
    ) as p:
        for _ in range(total_profile_steps):
            predictor.predict(data)
            torch.cuda.synchronize()
            p.step()


class InferenceBenchRunner(Runner):
    def __init__(
        self,
        run_dir_root,
        model_checkpoints: dict[str, str],
        natoms_list: list[int] | None = None,
        input_system: dict | None = None,
        timeiters: int = 10,
        seed: int = 1,
        device="cuda",
        overrides: dict | None = None,
        inference_settings: InferenceSettings = inference_settings_default(),  # noqa B008
        generate_traces: bool = False,  # takes additional memory and time
        dataset_name: str = "omat",
    ):
        self.natoms_list = natoms_list
        self.input_system = input_system
        assert (natoms_list is None or len(natoms_list) == 0) ^ (
            input_system is None
        ), "input must be either list of natoms or dict names: input system files"
        self.device = device
        self.seed = seed
        self.timeiters = timeiters
        self.model_checkpoints = model_checkpoints
        self.run_dir = os.path.join(run_dir_root, uuid.uuid4().hex.upper()[0:8])
        self.overrides = overrides
        self.inference_settings = inference_settings
        self.generate_traces = generate_traces
        self.dataset_name = dataset_name
        os.makedirs(self.run_dir, exist_ok=True)

    def run(self) -> None:
        seed_everywhere(self.seed)

        model_to_qps_data = defaultdict(list)

        for model_name, model_checkpoint in self.model_checkpoints.items():
            logging.info(
                f"Loading model: {model_checkpoint}, inference_settings: {self.inference_settings}"
            )
            predictor = MLIPPredictUnit(
                model_checkpoint,
                self.device,
                overrides=self.overrides,
                inference_settings=self.inference_settings,
            )
            max_neighbors = predictor.model.module.backbone.max_neighbors
            cutoff = predictor.model.module.backbone.cutoff
            logging.info(f"Model's max_neighbors: {max_neighbors}, cutoff: {cutoff}")

            # Bind loop-scoped variables as defaults to avoid late-binding (B023)
            def yield_inputs(_max_neighbors=max_neighbors, _cutoff=cutoff):
                if self.natoms_list is not None:
                    for natoms in self.natoms_list:
                        data = ase_to_graph(
                            get_fcc_carbon_xtal(natoms),
                            _max_neighbors,
                            _cutoff,
                            external_graph=self.inference_settings.external_graph_gen,
                            dataset_name=self.dataset_name,
                        )
                        yield data.natoms.item(), data
                else:
                    for k, v in self.input_system.items():
                        atoms = read(v)
                        if getattr(self, "expand_supercells", None) is not None:
                            size = self.expand_supercells
                            supercell_size = [[size, 0, 0], [0, size, 0], [0, 0, size]]
                            atoms = make_supercell(atoms, supercell_size)

                        data = ase_to_graph(
                            atoms,
                            _max_neighbors,
                            _cutoff,
                            external_graph=self.inference_settings.external_graph_gen,
                            dataset_name=self.dataset_name,
                        )
                        yield k, data

            # benchmark all models or number of atoms
            for name, data in yield_inputs():
                num_atoms = data.natoms.item()
                print_info = f"Starting profile: model: {model_checkpoint}, input: {name}, num_atoms: {num_atoms}"
                if self.inference_settings.external_graph_gen:
                    num_edges = data.edge_index.shape[1]
                    print_info += f" num edges compute on: {num_edges}"
                logging.info(print_info)
                if self.generate_traces:
                    make_profile(data, predictor, name=name, save_loc=self.run_dir)
                qps, ns_per_day = get_qps(data, predictor, timeiters=self.timeiters)
                model_to_qps_data[model_name].append([num_atoms, ns_per_day])
                logging.info(
                    f"Profile results: model: {model_checkpoint}, num_atoms: {num_atoms}, qps: {qps}, ns_per_day: {ns_per_day}"
                )

    def save_state(self, _):
        return

    def load_state(self, _):
        return



# ruff: noqa
from __future__ import annotations

import dataclasses
import json
import os
import shutil
import socket
import subprocess
import tempfile
import time
from typing import Callable, Optional, TypeVar
import uuid
from contextlib import closing
from pathlib import Path

import psutil
import submitit


def kill_proc_tree(pid, including_parent=True):
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    for child in children:
        child.kill()
    psutil.wait_procs(children, timeout=5)
    if including_parent:
        parent.kill()
        parent.wait(5)


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def scancel(job_ids: list[str]):
    """
    Cancel the SLURM jobs with the given job IDs.

    This function takes a list of job IDs.

    Args:
        job_ids (List[str]): A list of job IDs to cancel.
    """
    root_ids = list(set([i.split("_", maxsplit=2)[0] for i in job_ids]))
    subprocess.check_call(["scancel"] + root_ids)


start_ip_pattern = r"ray start --address='([0-9\.]+):([0-9]+)'"

PayloadReturnT = TypeVar("PayloadReturnT")


def mk_symlinks(target_dir: Path, job_type: str, paths: submitit.core.utils.JobPaths):
    """Create symlinks for the job's stdout and stderr in the target directory with a nicer name."""
    (target_dir / f"{job_type}.err").symlink_to(paths.stderr)
    (target_dir / f"{job_type}.out").symlink_to(paths.stdout)


@dataclasses.dataclass
class HeadInfo:
    """
    information about the head node that we can share to workers
    """

    hostname: Optional[str] = None
    port: Optional[int] = None
    temp_dir: Optional[str] = None


class RayClusterState:
    """
    This class is responsible for managing the state of the Ray cluster. It is useful to keep track
    of the head node and the workers, and to make sure they are all ready before starting the payload.

    It relies on storing info in a rendezvous directory so they can be shared async between jobs.

    Args:
        rdv_dir (Path): The directory where the rendezvous information will be stored. Defaults to ~/.fairray.
        cluster_id (str): A unique identifier for the cluster. Defaults to a random UUID. You only want to set this if you want to connect to an existing cluster.
    """

    def __init__(
        self,
        rdv_dir: Optional[Path] = None,
        cluster_id: Optional[str] = None,
    ):
        self.rendezvous_rootdir = (
            rdv_dir if rdv_dir is not None else (Path.home() / ".fairray")
        )
        self._cluster_id = (
            uuid.uuid4().hex if cluster_id is None else cluster_id
        )  # maybe use something more readable
        self.jobs_dir.mkdir(parents=True, exist_ok=True)

    @property
    def cluster_id(self) -> str:
        """Returns the unique identifier for the cluster."""
        return self._cluster_id

    @property
    def rendezvous_dir(self) -> Path:
        """Returns the path to the directory where the rendezvous information is stored."""
        return self.rendezvous_rootdir / self.cluster_id

    @property
    def jobs_dir(self) -> Path:
        """Returns the path to the directory where job information is stored."""
        return self.rendezvous_dir / "jobs"

    @property
    def _head_json(self) -> Path:
        """Returns the path to the JSON file containing head node information."""
        return self.rendezvous_dir / "head.json"

    def is_head_ready(self) -> bool:
        """Checks if the head node information is available and ready."""
        return self._head_json.exists()

    def head_info(self) -> Optional[HeadInfo]:
        """
        Retrieves the head node information from the stored JSON file.

        Returns:
            Optional[HeadInfo]: The head node information if available, otherwise None.
        """
        try:
            with self._head_json.open("r") as f:
                return HeadInfo(**json.load(f))
        except Exception as ex:
            print(f"failed to load head info: {ex}. Maybe it's not ready yet?")
            return None

    def save_head_info(self, head_info: HeadInfo):
        """
        Saves the head node information to a JSON file.

        Args:
            head_info (HeadInfo): The head node information to save.
        """
        with self._head_json.open("w") as f:
            json.dump(dataclasses.asdict(head_info), f)

    def clean(self):
        """Removes the rendezvous directory and all its contents."""
        shutil.rmtree(self.rendezvous_dir)

    def add_job(self, job: submitit.Job):
        """
        Adds a job to the jobs directory by creating a JSON file with the job's information.

        Args:
            job (submitit.Job): The job to add.
        """
        with (self.jobs_dir / f"{job.job_id}.json").open("w") as f:
            json.dump(
                {
                    "job_id": job.job_id,
                },
                fp=f,
            )

    def list_job_ids(self) -> list[str]:
        """Lists all job IDs stored in the jobs directory."""
        return [f.stem for f in self.jobs_dir.iterdir()]


def _ray_head_script(
    cluster_state: RayClusterState,
    worker_wait_timeout_seconds: int,
    payload: Optional[Callable[..., PayloadReturnT]] = None,
    **kwargs,
):
    """Start the head node of the Ray cluster on slurm."""
    hostname = socket.gethostname()
    head_env = os.environ.copy()
    num_cpus = os.environ.get("SLURM_CPUS_ON_NODE", 1)
    num_gpus = os.environ.get("SLURM_GPUS_ON_NODE", 0)
    # using 0 as the port for the head will make ray search for an open port instead of
    # always using the same one.
    port = find_free_port()
    head_env["RAY_ADDRESS"] = f"{hostname}:{port}"
    head_env["RAY_gcs_server_request_timeout_seconds"] = str(
        worker_wait_timeout_seconds
    )
    print(f"host {hostname}:{port}")
    with tempfile.TemporaryDirectory(dir="/tmp") as temp_dir:
        # ray workers have the same tempdir name (even on a different host)
        # as the head. This is a problem when we use  /scratch/slurm_tmpdir/JOBID as
        # the tempdir of the head job will not be accessible/visible from other workers if they
        # are scheduled on the same host. We are forced to use a different tempdir than /scratch
        # TODO ideally, we would still have a /scratch dir that everyone can share.
        process = subprocess.Popen(
            [
                "ray",
                "start",
                "--head",
                f"--port={port}",
                f"--temp-dir={temp_dir}",
                "--num-cpus",
                f"{num_cpus}",
                "--num-gpus",
                f"{num_gpus}",
                "--dashboard-host=0.0.0.0",
            ],
            env=head_env,
            stdout=subprocess.PIPE,
            text=True,
        )
        started = False
        for line in process.stdout:
            if "ray start --address=" in line:
                # this is a bit flaky, we search the stdout of the head job to
                # find this specific message and extract the address, it might be
                # better to not rely on ray printing this as it might change outside of our control.
                # Search for the pattern
                started = True
        assert (
            started
        ), "couldn't find head address in stdout. Check head.err for details"
        print(f"Head started, ip: {hostname}:{port} ({cluster_state.cluster_id})")
        info = HeadInfo(hostname=hostname, port=int(port), temp_dir=temp_dir)
        cluster_state.save_head_info(info)
        os.environ.update(head_env)
        if payload is not None:
            payload(**kwargs)
        else:
            while True:
                # practically, we should wait from driver signal to die here
                time.sleep(60)


def worker_script(
    cluster_state: RayClusterState,
    worker_wait_timeout_seconds: int,
    start_wait_time_seconds: int = 60,  # TODO pass this around properly
):
    """start an array of worker nodes for the Ray cluster on slurm. Waiting on the head node first."""
    print(f"Waiting for head node. {cluster_state.cluster_id}")
    while not cluster_state.is_head_ready():
        # wait for head to have started
        time.sleep(5)
    print("Head node found.")
    head_info = cluster_state.head_info()
    assert head_info is not None, "something went wrong getting head information."
    worker_env = os.environ.copy()
    worker_env["RAY_ADDRESS"] = f"{head_info.hostname}:{head_info.port}"
    worker_env["RAY_gcs_server_request_timeout_seconds"] = str(
        worker_wait_timeout_seconds
    )
    worker_env["RAY_raylet_start_wait_time_s"] = str(start_wait_time_seconds)
    num_cpus = os.environ.get("SLURM_CPUS_ON_NODE", 1)
    num_gpus = os.environ.get("SLURM_GPUS_ON_NODE", 0)

    try:
        subprocess.run(
            [
                "ray",
                "start",
                "--address",
                "auto",
                "--block",
                "--num-cpus",
                f"{num_cpus}",
                "--num-gpus",
                f"{num_gpus}",
            ],
            env=worker_env,
            check=False,
        )
    finally:
        if head_info.temp_dir:
            shutil.rmtree(Path(head_info.temp_dir))


# TODO deal with ports better: https://docs.ray.io/en/latest/cluster/vms/user-guides/community/slurm.html#slurm-networking-caveats
# TODO: reqs are just dicts, maybe we want to be more specific (in particular for qos/partition)
# TODO: need better naming too
# TODO: better log messages
# TODO checkpointing to recover worker nodes after timeout/preemption https://github.com/facebookincubator/submitit/blob/main/docs/checkpointing.md
# TODO have a ray autoscaler nodeprovider based on this, e.g. https://github.com/TingkaiLiu/Ray-SLURM-autoscaler/blob/main/slurm/node_provider.py
class RayCluster:
    """
    A RayCluster offers tools to start a Ray cluster (head and wokers) on slurm with the correct settings.

    args:

    log_dir: Path to the directory where logs will be stored. Defaults to "raycluster_logs" in the working directory. All slurm logs will go there,
    and it also creates symlinks to the stdout/stderr of each jobs with nicer name (head, worker_0, worker_1, ..., driver_0, etc). There interesting
    logs will be in the driver_N.err file, you should tail that.
    rdv_dir: Path to the directory where the rendezvous information will be stored. Defaults to ~/.fairray. Useful if you are trying to recover an existing cluster.
    cluster_id: A unique identifier for the cluster. Defaults to a random UUID. You only want to set this if you want to connect to an existing cluster.
    worker_wait_timeout_seconds (int): The number of seconds ray will wait for a worker to be ready before giving up. Defaults to 60 seconds. If you are scheduling
        workers in a queue that takes time for allocation, you might want to increase this otherwise your ray payload will fail, not finding resources.

    """

    log_dir: Path
    state: RayClusterState

    jobs: list[submitit.Job] = []
    is_shutdown = False
    num_worker_groups = 0
    num_drivers = 0
    head_started = False

    # keeping this in a separate object so it's easy to serialize and pass to jobs

    def __init__(
        self,
        log_dir: Path = Path("raycluster_logs"),
        rdv_dir: Optional[Path] = None,
        cluster_id: Optional[str] = None,
        worker_wait_timeout_seconds: int = 60,
    ):
        self.state = RayClusterState(rdv_dir, cluster_id)
        print(f"cluster {self.state.cluster_id}")
        self.log_dir = Path(log_dir) / self.state.cluster_id
        self.state.rendezvous_dir.mkdir(parents=True, exist_ok=True)
        self.worker_wait_timeout_seconds = worker_wait_timeout_seconds
        print(f"logs will be in {self.log_dir.resolve()}")

    def start_head(
        self,
        requirements: dict[str, int | str],
        name: str = "default",
        executor: str = "slurm",
        payload: Optional[Callable[..., PayloadReturnT]] = None,
        **kwargs,
    ) -> str:
        """
        Start the head node of the Ray cluster on slurm. You should do this first. Interesting requirements: qos, partition, time, gpus, cpus-per-task, mem-per-gpu, etc.
        """
        assert not self.head_started, "head already started"
        # start the head node
        self.head_started = True
        s_executor = submitit.AutoExecutor(
            folder=str(self.log_dir),
            cluster=executor,
        )
        s_executor.update_parameters(
            name=f"ray_head_{name}_{self.state.cluster_id}",  # TODO name should probably include more details (cluster_id)
            **requirements,
        )
        head_job = s_executor.submit(
            _ray_head_script,
            self.state,
            self.worker_wait_timeout_seconds,
            payload,
            **kwargs,
        )
        self.state.add_job(head_job)
        mk_symlinks(self.log_dir, "head", head_job.paths)
        print("head slurm job id:", head_job.job_id)
        return head_job.job_id

    def start_workers(
        self,
        num_workers: int,
        requirements: dict[str, int | str],
        name: str = "default",
        executor: str = "slurm",
    ) -> list[str]:
        """
        Start an array of worker nodes of the Ray cluster on slurm. You should do this after starting a head.
        Interesting requirements: qos, partition, time, gpus, cpus-per-task, mem-per-gpu, etc.
        You can call this multiple times to start an heterogeneous cluster.
        """
        # start the workers
        s_executor = submitit.AutoExecutor(folder=str(self.log_dir), cluster=executor)
        s_executor.update_parameters(
            name=f"ray_worker_{name}_{self.num_worker_groups}_{self.state.cluster_id}",  # TODO name should probably include more details (cluster_id)
            **requirements,
        )

        jobs = []
        with s_executor.batch():  # TODO set slurm array max parallelism here, because we really want all jobs to be scheduled at the same time
            for i in range(num_workers):
                jobs.append(
                    s_executor.submit(
                        worker_script,
                        self.state,
                        self.worker_wait_timeout_seconds,
                    )
                )

        for idx, j in enumerate(jobs):
            mk_symlinks(self.log_dir, f"worker_{self.num_worker_groups}_{idx}", j.paths)
        print("workers slurm job ids:", [job.job_id for job in jobs])
        for j in jobs:
            self.state.add_job(j)
        self.num_worker_groups += 1
        return [job.job_id for job in jobs]

    def shutdown(self):
        """
        Cancel all slurms jobs and get rid of rdv directory.
        """
        self.is_shutdown = True
        scancel(self.state.list_job_ids())
        kill_proc_tree(
            os.getpid(), including_parent=False
        )  # kill local job started by submitit as subprocess TODO that's not going to work when this is not the main process (e.g. recovering on cli)
        self.state.clean()
        print(f"cluster {self.state.cluster_id} shutdown")

    def __enter__(self):
        # only use as a context if you have something blocking waiting on the driver
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.shutdown()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import math
from contextlib import suppress
from dataclasses import dataclass

import torch
import torch.nn as nn
from torch.nn import functional as F

fairchem_cpp_found = False
with suppress(ModuleNotFoundError):
    import fairchem_cpp  # try to use DGL if available

    fairchem_cpp_found = True


def interval_intersection(interval1, interval2):
    """
    Compute intersection of two intervals [a, b] and [c, d]
    Returns None if no intersection, otherwise returns [start, end]
    """
    a, b = interval1
    c, d = interval2

    start = max(a, c)
    end = min(b, d)

    if start <= end:
        return [start, end]
    else:
        return None  # No intersection


def _softmax(x):
    return torch.softmax(x, dim=1) + 0.005


def _pnorm(x):
    return torch.nn.functional.normalize(x.abs() + 2 / x.shape[0], p=1.0, dim=1)


def norm_str_to_fn(act):
    if act == "softmax":
        return _softmax
    elif act == "pnorm":
        return _pnorm
    else:
        raise ValueError


@dataclass
class MOLEGlobals:
    # the linear coefficient for each expert
    expert_mixing_coefficients: torch.Tensor
    # if the input contains N separate systems, then the sizes represent the number of atoms in each system
    # this is used to for the MoLE to assign the correct parameters for each system
    mole_sizes: torch.Tensor
    # when using activation checkpointing, the inputs are chunked and given piecemeal so the start idx must be
    # updated each time the chunked operation happens. It's better to make this an input but in order for
    # the MolE interface to maintain functional equivalence to the Linear layer interface, this extra info
    # needs to be added here instead. (TODO: is there a cleaner way to do this?)
    ac_start_idx: int = 0


def init_linear(num_experts, use_bias, out_features, in_features):
    k = math.sqrt(1.0 / in_features)
    weights = nn.Parameter(
        k * 2 * (torch.rand(num_experts, out_features, in_features) - 0.5)
    )
    bias = nn.Parameter(k * 2 * (torch.rand(out_features) - 0.5)) if use_bias else None
    return weights, bias


class MOLEDGL(torch.nn.Module):
    def __init__(
        self,
        num_experts,
        in_features,
        out_features,
        global_mole_tensors,
        bias: bool,
    ):
        super().__init__()

        assert global_mole_tensors is not None
        self.num_experts = num_experts
        self.in_features = in_features
        self.out_features = out_features

        self.weights, self.bias = init_linear(
            num_experts, bias, out_features, in_features
        )

        self.global_mole_tensors = global_mole_tensors

    def forward(self, x):
        with torch.autocast(device_type=self.weights.device.type, enabled=False):
            weights = torch.einsum(
                "eoi, be->bio",
                self.weights,
                self.global_mole_tensors.expert_mixing_coefficients,
            )
        x_shape = x.shape
        if x.ndim == 2:
            r = fairchem_cpp.ops.segment_mm(
                x, weights, self.global_mole_tensors.mole_sizes
            )
        elif x.ndim == 3:
            r = fairchem_cpp.ops.segment_mm(
                x.reshape(-1, x_shape[-1]),
                weights,
                self.global_mole_tensors.mole_sizes * x_shape[1],
            ).reshape(*x_shape[:-1], -1)
        else:
            raise ValueError("x.ndim not in (2,3) not allowed")
        if self.bias is not None:
            r += self.bias
        return r


class MOLE(torch.nn.Module):
    def __init__(
        self,
        num_experts,
        in_features,
        out_features,
        global_mole_tensors: MOLEGlobals,
        bias: bool,
    ):
        super().__init__()
        self.num_experts = num_experts
        self.in_features = in_features
        self.out_features = out_features

        self.weights, self.bias = init_linear(
            num_experts, bias, out_features, in_features
        )

        self.global_mole_tensors = global_mole_tensors

    def merged_linear_layer(self):
        linear = torch.nn.Linear(
            in_features=self.in_features,
            out_features=self.out_features,
            bias=self.bias is not None,
        ).to(self.weights.device)

        with torch.autocast(device_type=self.weights.device.type, enabled=False):
            weights = torch.einsum(
                "eoi, be->boi",
                self.weights,
                self.global_mole_tensors.expert_mixing_coefficients,
            )

        with torch.no_grad():
            linear.weight.copy_(weights[0])
            if self.bias is not None:
                linear.bias.copy_(self.bias)
        return linear

    def forward(self, x):
        with torch.autocast(device_type=self.weights.device.type, enabled=False):
            weights = torch.einsum(
                "eoi, be->boi",
                self.weights,
                self.global_mole_tensors.expert_mixing_coefficients,
            )

        out = []
        ac_start_idx = self.global_mole_tensors.ac_start_idx
        assert len(self.global_mole_tensors.mole_sizes) > 0
        # TODO: precompute these if needed but they should be small and on cpu
        start_idxs = [0] + torch.cumsum(
            self.global_mole_tensors.mole_sizes, dim=0
        ).tolist()
        mole_intervals = list(zip(start_idxs, start_idxs[1:]))

        # Because activation checkpointing can chunk the inputs, we need to only compute
        # the mole_size intervals that overlap with the current chunks
        # for example if mole_sizes = [10,10,15]
        # start_idxs -> [0,10,20,35]
        # mole_intervals -> [(0,10),(10,20),(20,35)]
        # if the input segment is (5,15) then we compute the following 2 segments
        # (5,10),(10,15)
        input_segment = (ac_start_idx, ac_start_idx + x.shape[0])

        for n, mole_segment in enumerate(mole_intervals):
            interval_overlap = interval_intersection(input_segment, mole_segment)
            if interval_overlap is not None:
                start = interval_overlap[0] - ac_start_idx
                end = interval_overlap[1] - ac_start_idx
                out.append(F.linear(x[start:end], weights[n], bias=self.bias))

        result = torch.concatenate(out, dim=0)
        assert (
            result.shape[0] == x.shape[0]
        ), f"result shape {result.shape}, does not match input shape {x.shape} at dim 0"
        return result



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import bisect
from typing import TypeVar

import numpy as np
from omegaconf import DictConfig, OmegaConf
from torch.utils.data import Dataset

from fairchem.core.datasets.base_dataset import create_dataset

# This draws highly from fairseq's ConcatDataset implementation
# https://github.com/facebookresearch/fairseq/blob/nllb/fairseq/data/concat_dataset.py

T_co = TypeVar("T_co", covariant=True)


class ConcatDataset(Dataset[T_co]):
    @staticmethod
    def cumsum(sequence, sample_ratios):
        r, s = [], 0
        for e, ratio in zip(sequence, sample_ratios):
            curr_len = int(ratio * len(e))
            r.append(curr_len + s)
            s += curr_len
        return r

    def __init__(self, datasets, sampling: dict):
        super().__init__()
        if len(datasets) == 0:
            raise RuntimeError(
                "datasets should not be an empty iterable. at least one dataset must have 'key_mappings'"
            )

        # convert the dictionary into two same order lists
        self.datasets = []
        self.dataset_names = []
        for k, v in datasets.items():
            self.datasets.append(v)
            self.dataset_names.append(k)

        self.sample_ratios = self._dataset_sampling(
            [len(d) for d in self.datasets], self.dataset_names, sampling
        )
        self.cumulative_sizes = self.cumsum(self.datasets, self.sample_ratios)
        self.real_sizes = [len(d) for d in self.datasets]

    def __len__(self):
        return self.cumulative_sizes[-1]

    def __getitem__(self, idx):
        dataset_idx, sample_idx = self._get_dataset_and_sample_index(idx)
        data_object = self.datasets[dataset_idx][sample_idx]

        # Add additional attributes and cast to appropriate types
        # TODO setting data_object.dataset to dataset.split breaks the collater due to the task map
        #  we should clean that up by using a .dataset and .split keys for all data objects
        data_object.dataset_name = self.dataset_names[dataset_idx]
        # TODO this should be ensured in the datasets themselves, rather than casting/redifining here?
        data_object.fixed = data_object.fixed.long()
        data_object.atomic_numbers = data_object.atomic_numbers.long()

        return data_object

    def _get_dataset_and_sample_index_list(self, sample_idxs: list):
        sample_idxs = np.array(sample_idxs)

        # find out which dataset owns which sample_idx and what the internal idx is for each
        dataset_idx_ownership = np.zeros(sample_idxs.shape[0], dtype=np.int64)
        internal_sample_idxs = np.zeros(sample_idxs.shape[0], dtype=np.int64)
        for dataset_idx in range(len(self.cumulative_sizes)):
            # find out what dataset owns this idx
            this_dataset_idx_ownership = np.where(
                sample_idxs < self.cumulative_sizes[dataset_idx]
            )[0]
            dataset_idx_ownership[this_dataset_idx_ownership] = dataset_idx

            # figure out the offset needed to subtract to get internal index
            offset = 0
            if dataset_idx > 0:
                offset = self.cumulative_sizes[dataset_idx - 1]

            # update internal idx
            internal_sample_idxs[this_dataset_idx_ownership] = (
                sample_idxs[this_dataset_idx_ownership] - offset
            ) % self.real_sizes[dataset_idx]

            # set the just consumed idxs to out of bounds
            sample_idxs[this_dataset_idx_ownership] = self.cumulative_sizes[-1] + 1

        return dataset_idx_ownership, internal_sample_idxs

    # @functools.cache
    # B019 Use of `functools.lru_cache` or `functools.cache` on methods can lead to memory leaks
    def _get_dataset_and_sample_index(self, idx: int):
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        sample_idx = sample_idx % self.real_sizes[dataset_idx]
        return dataset_idx, sample_idx

    @property
    def updated_dataset_sizes(self):
        new_sizes = [
            (
                self.cumulative_sizes[i]
                if i == 0
                else self.cumulative_sizes[i] - self.cumulative_sizes[i - 1]
            )
            for i, _ in enumerate(self.cumulative_sizes)
        ]
        return (self.real_sizes, new_sizes, self.sample_ratios)

    def metadata_hasattr(self, attr) -> bool:
        for dataset in self.datasets:  # noqa: SIM110
            if not dataset.metadata_hasattr(attr):
                return False
        return True

    def get_metadata(self, attr, sample_idxs_to_get_metadata_for):
        assert attr == "natoms"
        if isinstance(sample_idxs_to_get_metadata_for, list):
            metadata = np.zeros(len(sample_idxs_to_get_metadata_for), dtype=np.int32)
            dataset_idxs, dataset_internal_sample_idx = (
                self._get_dataset_and_sample_index_list(sample_idxs_to_get_metadata_for)
            )
            for dataset_idx in range(len(self.cumulative_sizes)):
                dataset_mask = dataset_idxs == dataset_idx
                metadata[dataset_mask] = self.datasets[dataset_idx].get_metadata(
                    "natoms", list(dataset_internal_sample_idx[dataset_mask])
                )[0]
            return metadata
        dataset_idx, sample_idx = self._get_dataset_and_sample_index(
            sample_idxs_to_get_metadata_for
        )
        return self.datasets[dataset_idx].get_metadata("natoms", sample_idx)

    @staticmethod
    def _dataset_sampling(
        dataset_sizes: list[int], dataset_names: list[str], sampling: dict
    ) -> list[float]:
        """
        Return expansion ratios for each dataset based on sampling strategy
        """
        if sampling["type"] == "explicit":
            for dataset_name in dataset_names:
                if dataset_name not in sampling["ratios"]:
                    raise ValueError(
                        f"Missing ratio for dataset with name: {dataset_name}"
                    )
            return [sampling["ratios"][dataset_name] for dataset_name in dataset_names]
        if sampling["type"] == "balanced":
            indv_target_size = max(dataset_sizes)
            return [indv_target_size / size for size in dataset_sizes]
        elif sampling["type"] == "temperature":
            assert (
                sampling["temperature"] >= 1.0
            ), "Temperature must be >= 1.0, for custom weights use weighted sampling."
            # total size before expansion
            total_size = sum(dataset_sizes)
            # calc the temperature sampling probabilities for each dataset
            # p_i = (D_i / D_total)^(1/T)
            temp_prob = [
                (size / total_size) ** (1.0 / sampling["temperature"])
                for size in dataset_sizes
            ]
            # normalize temp probabilities i.e. sum to 1
            temp_prob = [r / sum(temp_prob) for r in temp_prob]
            # find the target dataset size with temp sampling
            # this assume largest dataset ratio is 1
            max_idx = np.argmax(dataset_sizes)
            target_size = dataset_sizes[max_idx] / temp_prob[max_idx]
            # return expansion ratios
            ratios = (target_size * np.array(temp_prob)) / np.array(dataset_sizes)
            return ratios.tolist()
        elif sampling["type"] == "weighted":
            return sampling["ratios"]
        else:
            raise NotImplementedError(f"{sampling} not implemented.")


# This version is used for the Hydra configs
def create_concat_dataset(
    dataset_configs: DictConfig, combined_dataset_config: dict
) -> ConcatDataset:
    """Make a concat dataset with all the splits for each dataset. Keys will be {dataset}.{split}"""
    datasets: dict[str, Dataset] = {}
    # we need to convert omegaConf DictConfig objects to a straight up dict
    dataset_configs = OmegaConf.to_object(dataset_configs)
    for dataset_name in sorted(dataset_configs.keys()):
        dataset_config = dataset_configs[dataset_name]
        assert (
            dataset_config.get("lin_ref", None) is None
        ), "lin_refs in the dataset config are deprecated, please move them to the task config"
        for split in dataset_config["splits"]:
            try:
                datasets[f"{dataset_name}.{split}"] = create_dataset(
                    dataset_config, split=split
                )
            except ValueError as e:
                raise ValueError(f"Error creating dataset '{dataset_name}' {e}") from e

    return ConcatDataset(datasets, sampling=combined_dataset_config["sampling"])



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from numbers import Number
from typing import TYPE_CHECKING, Protocol, overload, runtime_checkable

if TYPE_CHECKING:
    from ase import Atoms

    from fairchem.core.datasets.ase_datasets import AseAtomsDataset


@runtime_checkable
class AtomsSequence(Protocol):
    @overload
    def __getitem__(self, index: int) -> Atoms: ...

    @overload
    def __getitem__(self, index: slice) -> AtomsSequence: ...

    def __getitem__(self, index) -> Atoms | AtomsSequence: ...

    def __len__(self) -> int: ...


class AtomsDatasetSequence:
    """
    Turn an AseAtomsDataset into an AtomsSequence that iterates over atoms objects.
    """

    def __init__(self, dataset: AseAtomsDataset):
        self.dataset = dataset

    def __getitem__(self, index: int | slice) -> Atoms | AtomsSequence:
        if isinstance(index, Number):
            return self.dataset.get_atoms(index)
        else:
            raise IndexError("Unsupported indexing")

    def __len__(self) -> int:
        return len(self.dataset)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
import traceback
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Sequence

import numpy as np
import pandas as pd
from monty.dev import requires
from tqdm import tqdm

from fairchem.core.components.calculate._calculate_runner import CalculateRunner
from fairchem.core.components.calculate.recipes.phonons import (
    run_mdr_phonon_benchmark,
)

try:
    import phonopy

    phonopy_installed = True
except ImportError:
    phonopy_installed = False


if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator


def get_mdr_phonon_data_list(index_df_path, phonon_file_path, debug=False):
    ref_df = pd.read_json(index_df_path)
    if debug:
        ref_df = ref_df[ref_df["nsites"] <= 5].head(10)
    all_ids = list(ref_df["mp_id"])
    phonon_yamls = [Path(phonon_file_path) / f"{mpid}.yaml.bz2" for mpid in all_ids]
    return phonon_yamls


@requires(phonopy_installed, message="Requires `phonopy` to be installed")
class MDRPhononRunner(CalculateRunner):
    """Calculate elastic tensor for a set of structures."""

    result_glob_pattern: ClassVar[str] = "mdr_phonon_dist*_*-*.json.gz"

    def __init__(
        self,
        calculator: Calculator,
        input_data: Sequence[dict],
        displacement: float = 0.01,
    ):
        """Initialize the CalculateRunner with a calculator and input data.

        Args:
            calculator (Calculator): ASE-like calculator to perform calculations
            input_data (Sequence): Input data to be processed by the calculator
        """
        super().__init__(calculator, input_data)
        self.displacement = displacement

    # TODO continue if unfinished
    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> list[dict[str, Any]]:
        all_results = []
        chunk_indices = np.array_split(range(len(self.input_data)), num_jobs)[job_num]
        for i in tqdm(chunk_indices, desc="Running phonon calculations."):
            try:
                phonon_file_name = self.input_data[i]
                phonon = phonopy.load(phonon_file_name)
                phonon_results = run_mdr_phonon_benchmark(
                    phonon,
                    calculator=self.calculator,
                    displacement=self.displacement,
                    run_relax=True,
                    fix_symm_relax=False,
                    symprec=1e-4,
                    symmetrize_fc=False,
                )

                frequencies = phonon_results["frequencies"]
                max_freq = frequencies.max()
                avg_freq = frequencies.mean()
                # temp at 300 K
                entropy = phonon_results["entropy"][4]
                heat_capacity = phonon_results["heat_capacity"][4]
                free_energy = phonon_results["free_energy"][4]

                results = {
                    "sid": phonon_file_name.stem.split(".")[0],
                    "energy_per_atom": phonon_results["energy_per_atom"],
                    "volume_per_atom": phonon_results["volume_per_atom"],
                    "max_freq": max_freq,
                    "avg_freq": avg_freq,
                    "entropy": entropy,
                    "heat_capacity": heat_capacity,
                    "free_energy": free_energy,
                    "errors": "",
                    "traceback": "",
                }

            except Exception as ex:
                results = {
                    "sid": phonon_file_name.stem.split(".")[0],
                    "energy_per_atom": np.nan,
                    "volume_per_atom": np.nan,
                    "max_freq": np.nan,
                    "avg_freq": np.nan,
                    "entropy": np.nan,
                    "heat_capacity": np.nan,
                    "free_energy": np.nan,
                    "errors": (f"{ex!r}"),
                    "traceback": (traceback.format_exc()),
                }

            all_results.append(results)

        return all_results

    def write_results(
        self,
        results: list[dict[str, Any]],
        results_dir: str,
        job_num: int = 0,
        num_jobs: int = 1,
    ) -> None:
        results_df = pd.DataFrame(results)
        results_df.to_json(
            os.path.join(
                results_dir,
                f"mdr_phonon_dist{self.displacement}_{num_jobs}-{job_num}.json.gz",
            )
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ase.calculators.calculator import PropertyNotPresent

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ase import Atoms
    from numpy.typing import ArrayLike


def get_property_from_atoms(
    atoms: Atoms, property_name: str
) -> int | float | ArrayLike:
    """Retrieve a property from an Atoms object, either from its properties or info dictionary.

    Args:
        atoms: The ASE Atoms object to extract properties from
        property_name: Name of the property to retrieve

    Returns:
        The property value as an integer, float, or array-like object

    Raises:
        ValueError: If the property is not found in either the properties or info dictionary
    """
    try:
        # get_properties returns a Properties dict-like object, so we index again for the property requested
        prop = atoms.get_properties([property_name])[property_name]
    except (PropertyNotPresent, ValueError, RuntimeError):
        try:
            prop = atoms.info[property_name]
        except KeyError as err:
            raise ValueError(
                f"The listed property {property_name} in `save_target_properties` is not available from"
                f" the atoms object or its info dictionary"
            ) from err
    return prop


def normalize_property(
    property_value: float | ArrayLike, atoms: Atoms, normalize_by: str
):
    """Normalize a property value by either the number of atoms or another property.

    Args:
        property_value: The property value to normalize
        atoms: The ASE Atoms object containing the normalization information
        normalize_by: Normalization method, either "natoms" to divide by number of atoms
            or a property name to divide by that property's value

    Returns: The normalized property value
    """
    if normalize_by == "natoms":
        return property_value / len(atoms)
    else:
        norm_prop = get_property_from_atoms(atoms, normalize_by)
        return property_value / norm_prop


def get_property_dict_from_atoms(
    properties: Sequence[str], atoms: Atoms, normalize_by: dict[str, str] | None = None
) -> dict[str, float | ArrayLike]:
    """Get a sequence of properties from an atoms object and return a dict.

    Args:
        properties: Sequence of property names to retrieve from the atoms object
        atoms: The ASE Atoms object to extract properties from
        normalize_by: Dictionary mapping property names to normalization methods

    Returns:
        Dictionary containing the requested properties as keys and
            normalized properties if specified in normalize_by
    """
    normalize_by = normalize_by or {}
    results = {}
    for property_name in properties:
        results[property_name] = get_property_from_atoms(atoms, property_name)
        if property_name in normalize_by:
            norm_by = normalize_by[property_name]
            key = (
                f"{property_name}_per_atom"
                if norm_by == "natoms"
                else f"{property_name}_per_{norm_by}"
            )
            results[key] = normalize_property(results[property_name], atoms, norm_by)
    return results



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse
import os
import pickle

import omegaconf
import torch

from fairchem.core.scripts.migrate_imports import mapping
from fairchem.core.units.mlip_unit import MLIPPredictUnit


def find_new_module_name(module):
    if module in mapping:
        return mapping[module]
    module_path = module.split(".")
    # check if parent resolves
    parent_module = ".".join(module_path[:-1])
    if parent_module == "":
        return module_path[0]
    else:
        return f"{find_new_module_name(parent_module)}.{module_path[-1]}"


def update_config(config_or_data):
    if isinstance(config_or_data, (omegaconf.dictconfig.DictConfig, dict)):
        for k, v in config_or_data.items():
            config_or_data[k] = update_config(v)
    elif isinstance(config_or_data, (omegaconf.listconfig.ListConfig, list)):
        for i, item in enumerate(config_or_data):
            config_or_data[i] = update_config(item)
    elif isinstance(config_or_data, str):
        for k, v in mapping.items():
            if k in config_or_data:
                config_or_data = config_or_data.replace(k, v)
        config_or_data = config_or_data.replace("osc", "omc")
    return config_or_data


class RenameUnpickler(pickle.Unpickler):
    def __init__(self, file, *args, **kwargs):
        super().__init__(file, *args, **kwargs)

    def find_class(self, module, name):
        return super().find_class(find_new_module_name(module), name)


def generate_stress_task_config(dataset_name, task_name, rmsd):
    return {
        "_target_": "fairchem.core.units.mlip_unit.mlip_unit.Task",
        "name": task_name,
        "level": "system",
        "property": "stress",
        "loss_fn": {
            "_target_": "fairchem.core.modules.loss.DDPMTLoss",
            "loss_fn": {"_target_": "fairchem.core.modules.loss.MAELoss"},
            "reduction": "mean",
            "coefficient": 1,
        },
        "out_spec": {"dim": [1, 9], "dtype": "float32"},
        "normalizer": {
            "_target_": "fairchem.core.modules.normalization.normalizer.Normalizer",
            "mean": 0.0,
            "rmsd": rmsd,  # 1.423
        },
        "datasets": [dataset_name],
        "metrics": ["mae"],
    }


def migrate_checkpoint(
    checkpoint_path: torch.nn.Module,
    rm_static_keys: bool = True,
    map_undefined_stress_to: str | None = None,
    add_stress: bool = False,
    task_add_stress: str | None = None,
    model_version: float = 1.0,
) -> dict:
    """
    Migrates a checkpoint by updating module imports and configurations.

    This function loads a checkpoint, updates its configuration using the mapping
    defined in fairchem.core.scripts.migrate_imports,

    optionally adds stress tasks for datasets that don't have them,
    and optionally removes static keys that are no longer needed.

    Args:
        checkpoint_path: Path to the input checkpoint file
        rm_static_keys: Whether to remove static keys from the state dictionaries
        task_add_stress: If provided, adds stress tasks for datasets based on this task
        model_version: Inject this model version into model

    Returns:
        Migrated checkpoint dict
    """
    pickle.Unpickler = RenameUnpickler
    checkpoint = torch.load(checkpoint_path, pickle_module=pickle)
    checkpoint.tasks_config = update_config(checkpoint.tasks_config)
    checkpoint.model_config = update_config(checkpoint.model_config)

    if (
        checkpoint.model_config["backbone"]["model"]
        == "fairchem.experimental.foundation_models.models.message_passing.escn_omol.eSCNMDBackbone"
    ):
        checkpoint.model_config["backbone"]["use_dataset_embedding"] = False

    if (
        checkpoint.model_config["backbone"]["model"]
        == "fairchem.core.models.uma.escn_moe.eSCNMDMoeBackbone"
    ):
        if "model_version" in checkpoint.model_config["backbone"]:
            assert checkpoint.model_config["backbone"]["model_version"] == model_version
        print("Setting model version to", model_version)
        checkpoint.model_config["backbone"]["model_version"] = model_version

    if task_add_stress is not None:
        target_stress_task = f"{task_add_stress}_stress"
        output_dataset_names = set()
        datasets_with_stress = set()
        target_stress_config = None
        # find output datasets
        for task in checkpoint.tasks_config:
            if "_energy" in task.name:
                output_dataset_names.add(task.name.replace("_energy", ""))
            elif "_stress" in task.name:
                datasets_with_stress.add(task.name.replace("_stress", ""))
                if task.name == target_stress_task:
                    target_stress_config = task
        assert (
            target_stress_config is not None
        ), f"Did not find existing task {target_stress_task} in {[task.name for task in checkpoint.tasks_config]}"
        # copy over the task configs to tasks that dont have stress
        for dataset_name in output_dataset_names - datasets_with_stress:
            task_config = target_stress_config.copy()
            task_config.name = f"{dataset_name}_stress"
            task_config.datasets = [dataset_name]
            checkpoint.tasks_config.append(task_config)

    if add_stress:
        checkpoint.model_config["backbone"]["regress_stress"] = True
        output_dataset_names = set()
        datasets_with_stress = set()
        datasets_to_rmsd = {}
        # find output datasets

        # figure out if this is a UMA like model
        uma_like_model = any(
            "_energy" in task.name and "omol" in task.name
            for task in checkpoint.tasks_config
        )
        if uma_like_model:
            # we dont have tests here , preserving this for uma like model migration
            for task in checkpoint.tasks_config:
                if "_energy" in task.name:
                    output_dataset_names.add(task.name.replace("_energy", ""))
                    datasets_to_rmsd[task.name.replace("_energy", "")] = task[
                        "normalizer"
                    ]["rmsd"]
            for dataset_name in output_dataset_names - datasets_with_stress:
                checkpoint.tasks_config.append(
                    generate_stress_task_config(
                        dataset_name,
                        f"{dataset_name}_stress",
                        datasets_to_rmsd[dataset_name],
                    )
                )
        else:
            energy_tasks = list(
                filter(lambda task: task.name == "energy", checkpoint.tasks_config)
            )
            assert (
                len(energy_tasks) == 1
            ), f"Expected exactly one energy task for non uma like model, found {[task.name for task in energy_tasks]}"
            energy_task = energy_tasks[0]
            assert (
                len(energy_task.datasets) == 1
            ), f"Expected exactly one dataset for energy task, found {energy_task.datasets}"
            checkpoint.tasks_config.append(
                generate_stress_task_config(
                    energy_task.datasets[0], "stress", energy_task["normalizer"]["rmsd"]
                )
            )

    # remove keys for registered buffers that are no longer saved
    if rm_static_keys:
        remove_keys = {"expand_index", "offset", "balance_degree_weight"}
        # list explicit keys to rename here in the weight state dictionary
        rename_keys = {
            # "module.backbone.routing_mlp": "module.backbone.mole_coefficient_mlp",
            # "module.backbone.moe_coefficient_mlp": "module.backbone.mole_coefficient_mlp",
            "backbone.dataset_embedding.dataset_emb_dict.osc.weight": "backbone.dataset_embedding.dataset_emb_dict.omc.weight",
            "module.backbone.dataset_embedding.dataset_emb_dict.osc.weight": "module.backbone.dataset_embedding.dataset_emb_dict.omc.weight",
        }
        for state_dict_name in ["model_state_dict", "ema_state_dict"]:
            state_dict = getattr(checkpoint, state_dict_name)
            for k in [key for key in state_dict if key.split(".")[-1] in remove_keys]:
                state_dict.pop(k)
                print(f"Removing {k} from {state_dict_name}")
            # rename explicitly mapped keys
            for subkey_from, subkey_to in rename_keys.items():
                for key_from in [key for key in state_dict if subkey_from in key]:
                    key_to = key_from.replace(subkey_from, subkey_to)
                    state_dict[key_to] = state_dict[key_from]
                    state_dict.pop(key_from)
                    print("rename", key_from, key_to)

    return checkpoint


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert ocp_dev/foundation_model checkpoint to fm_release"
    )

    parser.add_argument(
        "--checkpoint-in", type=str, help="checkpoint input", required=True
    )
    parser.add_argument(
        "--checkpoint-out", type=str, help="checkpoint output", required=True
    )
    parser.add_argument(
        "--remove-static-keys", default=True, action=argparse.BooleanOptionalAction
    )
    parser.add_argument(
        "--add-stress", default=False, action=argparse.BooleanOptionalAction
    )
    parser.add_argument(
        "--map-undefined-stress-to", type=str, required=False, default=None
    )
    parser.add_argument("--model-version", type=float, default=1.0)
    args = parser.parse_args()

    if os.path.exists(args.checkpoint_out):
        raise FileExistsError(
            f"Output checkpoint ({args.checkpoint_out}) cannot already exist"
        )

    checkpoint = migrate_checkpoint(
        args.checkpoint_in,
        args.remove_static_keys,
        args.map_undefined_stress_to,
        add_stress=args.add_stress,
        model_version=args.model_version,
    )
    torch.save(checkpoint, args.checkpoint_out)

    # test to see if checkpoint loads
    MLIPPredictUnit(args.checkpoint_out, device="cpu")

    # task_name="omol"
    # calc = FAIRChemCalculator.from_model_checkpoint(args.checkpoint_out,task_name)
    # atoms = molecule("H2O")
    # atoms.set_cell([100.0, 100.0, 100.0])  # Define a cubic cell
    # atoms.set_pbc(False)  # Enable periodic boundary conditions
    # atoms.calc = calc
    # energy = atoms.get_potential_energy()
    # stress=atoms.get_stress(task_name)
